#define _GNU_SOURCE

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

#include "rdtsc.h"

const int niterations = 100000;

#define TEN_TIMES(x) x x x x x x x x x x
#define THOUSAND_TIMES(x) TEN_TIMES(TEN_TIMES(TEN_TIMES(x)))

volatile int a1, a2, a3, a4, a5, a6, a7, a8, a9;

#define ONE_WRITE                 a1 = 1;
#define TWO_WRITES   ONE_WRITE    a2 = 2;
#define THREE_WRITES TWO_WRITES   a3 = 3;
#define FOUR_WRITES  THREE_WRITES a4 = 4;
#define FIVE_WRITES  FOUR_WRITES  a5 = 5;
#define SIX_WRITES   FIVE_WRITES  a6 = 6;
#define SEVEN_WRITES SIX_WRITES   a7 = 7;
#define EIGHT_WRITES SEVEN_WRITES a8 = 8;
#define NINE_WRITES  EIGHT_WRITES a9 = 9;
#define WRITES_COUNT 10

#define MFENCE __asm("mfence");
#define LOCK_ADD __asm("lock addl $0,(%rsp)");

#define BENCHMARK(result, what) { 					\
	uint64_t ticks_total = 0;					\
									\
	for (int i = 0; i < niterations; ++i) {				\
		uint64_t begin = rdtsc();				\
		THOUSAND_TIMES(what)					\
		uint64_t end = rdtsc();					\
									\
		ticks_total += end - begin;				\
	}								\
									\
	if (my_id == 0) {						\
		result = ticks_total;					\
	}								\
}

uint64_t mfence_timings[WRITES_COUNT];
uint64_t lock_add_timings[WRITES_COUNT];
uint64_t writes_timings[WRITES_COUNT];

void *thread_main(void *ptr) {
	int my_id = *(int *)ptr;

	cpu_set_t cpu_set;

	CPU_ZERO(&cpu_set);
	CPU_SET(my_id, &cpu_set);

	if (pthread_setaffinity_np(pthread_self(), sizeof(cpu_set), &cpu_set) != 0) {
		perror("sched_setaffinity");
		return NULL;
	}

	BENCHMARK(mfence_timings[0], MFENCE)
	BENCHMARK(mfence_timings[1], ONE_WRITE    MFENCE)
	BENCHMARK(mfence_timings[2], TWO_WRITES   MFENCE)
	BENCHMARK(mfence_timings[3], THREE_WRITES MFENCE)
	BENCHMARK(mfence_timings[4], FOUR_WRITES  MFENCE)
	BENCHMARK(mfence_timings[5], FIVE_WRITES  MFENCE)
	BENCHMARK(mfence_timings[6], SIX_WRITES   MFENCE)
	BENCHMARK(mfence_timings[7], SEVEN_WRITES MFENCE)
	BENCHMARK(mfence_timings[8], EIGHT_WRITES MFENCE)
	BENCHMARK(mfence_timings[9], NINE_WRITES  MFENCE)

	BENCHMARK(lock_add_timings[0], LOCK_ADD)
	BENCHMARK(lock_add_timings[1], ONE_WRITE    LOCK_ADD)
	BENCHMARK(lock_add_timings[2], TWO_WRITES   LOCK_ADD)
	BENCHMARK(lock_add_timings[3], THREE_WRITES LOCK_ADD)
	BENCHMARK(lock_add_timings[4], FOUR_WRITES  LOCK_ADD)
	BENCHMARK(lock_add_timings[5], FIVE_WRITES  LOCK_ADD)
	BENCHMARK(lock_add_timings[6], SIX_WRITES   LOCK_ADD)
	BENCHMARK(lock_add_timings[7], SEVEN_WRITES LOCK_ADD)
	BENCHMARK(lock_add_timings[8], EIGHT_WRITES LOCK_ADD)
	BENCHMARK(lock_add_timings[9], NINE_WRITES  LOCK_ADD)

	BENCHMARK(writes_timings[0], )
	BENCHMARK(writes_timings[1], ONE_WRITE)
	BENCHMARK(writes_timings[2], TWO_WRITES)
	BENCHMARK(writes_timings[3], THREE_WRITES)
	BENCHMARK(writes_timings[4], FOUR_WRITES)
	BENCHMARK(writes_timings[5], FIVE_WRITES)
	BENCHMARK(writes_timings[6], SIX_WRITES)
	BENCHMARK(writes_timings[7], SEVEN_WRITES)
	BENCHMARK(writes_timings[8], EIGHT_WRITES)
	BENCHMARK(writes_timings[9], NINE_WRITES)

	return NULL;
}

int main(int argc, char *argv[]) {
	if (argc != 2) {
		fprintf(stderr, "Usage: rdtsc-mfence NTHREADS\n");
		return 1;
	}

	errno = 0;
	int threads_count = strtol(argv[1], NULL, 10);

	if (errno != 0) {
		fprintf(stderr, "%s is not a valid integer", argv[1]);
		return 1;
	}
	if (threads_count < 1) {
		fprintf(stderr, "number of threads must be at least one");
		return 1;
	}


	int thread_ids[threads_count];
	pthread_t threads[threads_count];

	for (int i = 0; i < threads_count; ++i) {
		thread_ids[i] = i;
		if (pthread_create(&threads[i], 0, thread_main, &thread_ids[i]) != 0) {
			perror("pthread_create");
			return 1;
		}
	}

	for (int i = 0; i < threads_count; ++i) {
		if (pthread_join(threads[i], NULL) != 0) {
			perror("pthread_join");
			return 1;
		}
	}

	for (int i = 0; i < WRITES_COUNT; ++i) {
		printf("%d %lf %lf %lf\n", i,
			0.001 * (mfence_timings[i] / niterations),
			0.001 * (lock_add_timings[i] / niterations),
			0.001 * (writes_timings[i] / niterations));
	}

	return 0;
}
